@@ -5,7 +5,7 @@
  *
  * KSBA is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2 of the License, or
+ * the Free Software Foundation; either version 3 of the License, or
  * (at your option) any later version.
  *
  * KSBA is distributed in the hope that it will be useful,
@@ -14,8 +14,7 @@
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
+ * along with this program; if not, see <http://www.gnu.org/licenses/>.
  */
 
 #ifndef UTIL_H
@@ -49,10 +48,11 @@ char *_ksba_xstrdup (const char *p);
 #define DIM(v) (sizeof(v)/sizeof((v)[0]))
 #define DIMof(type,member)   DIM(((type *)0)->member)
 #ifndef STR
-  #define STR(v) #v
+# define STR(v) #v
+#endif
+#ifndef STR2
+# define STR2(v) STR(v)
 #endif
-#define STR2(v) STR(v)
-
 
 #define return_if_fail(expr) do {                        \
     if (!(expr)) {                                       \
@@ -99,9 +99,3 @@ char *_ksba_stpcpy (char *a, const char *b);
 
 #endif /* UTIL_H */
 
-
-
-
-
-
-
